#include <stdint.h>
#include "dmf_controller.h"

#ifndef AVR
using namespace std;
#include <boost/date_time/posix_time/posix_time_types.hpp>
#else
#include "Arduino.h" //for arduino version > 1.0 it is arduino.h and < 1.0 is WProgram.h
#include <Wire.h>
#include <SPI.h>
#include <math.h>
#include <Servo.h>

// map the electrodes on the DMF device to the ports on the MAX7300
const uint8_t DmfController::MAX7300_PORT_ADDRESSES[] = {
                              0x28,0x2C,0x29,0x2D,0x2A,0x2E,0x2B,0x2F,0x30,0x31,
                              //0    1    2    3    4    5    6    7    8    9
                              0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3A,0x3B,
                              //10   11   12   13   14   15   16   17   18   19
                              0x3C,0x27,0x3D,0x26,0x3E,0x25,0x3F,0x24};
                              //20   21   22   23   24   25   26   27

#endif

const char DmfController::PROTOCOL_NAME_[] =
  "Digital Microfluidics Control Protocol";
const char DmfController::PROTOCOL_VERSION_[] = "0.1";
const char DmfController::VERSION_[] = "0.1";
#ifdef AVR
const char DmfController::NAME_[] = "Arduino DMF Controller";
#else
const char DmfController::NAME_[] = "DmfController C++ Class";
const char DmfController::CSV_INDENT_[] = ",,,,,,,,";
#endif


DmfController::DmfController()
  : SerialProtocol(BAUD_RATE,false
#ifndef AVR
                   ,"DmfController" //used for logging
#endif
                   ) {
}

DmfController::~DmfController() {
}


void DmfController::ProcessCommand(uint8_t cmd) {
#ifndef AVR
  const char* function_name = "ProcessCommand()";
  sprintf(log_message_string_,"command=0x%0X (%d)",
          cmd,cmd);
  LogMessage(log_message_string_, function_name);
#endif
  uint8_t return_code = RETURN_UNKNOWN_COMMAND;
  switch(cmd) {
    case CMD_GET_DEVICE_NAME:
#ifndef AVR
      LogMessage("CMD_GET_DEVICE_NAME", function_name);
#endif
      if(payload_length()==0) {
        Serialize(NAME_,sizeof(NAME_));
        return_code = RETURN_OK;
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_GET_DEVICE_VERSION:
#ifndef AVR
      LogMessage("CMD_GET_DEVICE_VERSION", function_name);
#endif
      if(payload_length()==0) {
        Serialize(VERSION_,sizeof(VERSION_));
        return_code = RETURN_OK;
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
#ifdef AVR // Commands that only the Arduino handles
    case CMD_GET_NUMBER_OF_ELECTRODES:
      if(payload_length()==0) {
        uint16_t n = NUMBER_OF_ELECTRODES; // compiler complains when we pass a
        Serialize(&n,sizeof(n));           // static const
        return_code = RETURN_OK;
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_GET_STATE_OF_ALL_ELECTRODES:
      if(payload_length()==0) {
        Serialize(state_of_electrodes_,NUMBER_OF_ELECTRODES*sizeof(uint8_t));
        return_code = RETURN_OK;
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_SET_STATE_OF_ALL_ELECTRODES:
      if(payload_length()==NUMBER_OF_ELECTRODES*sizeof(uint8_t)) {
        ReadArray(state_of_electrodes_,
                  NUMBER_OF_ELECTRODES*sizeof(uint8_t));
        UpdateAllElectrodes();
        return_code = RETURN_OK;
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_SET_STATE_OF_ELECTRODE:
      if(payload_length()==sizeof(uint16_t)+sizeof(uint8_t)) {
        uint16_t electrode = ReadUint16();
        if(electrode<NUMBER_OF_ELECTRODES) {
          state_of_electrodes_[electrode] = ReadUint8();
          UpdateElectrode(electrode);
          return_code = RETURN_OK;
        } else {
          return_code = RETURN_BAD_INDEX;
        }
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_SET_ACTUATION_VOLTAGE:
      if(payload_length()==sizeof(float)) {
        float voltage = ReadFloat();
        if(voltage/AMPLIFIER_GAIN>5) {
          return_code = RETURN_GENERAL_ERROR;
        } else {
          SendSPI(AD5206_SLAVE_SELECT_PIN,4,
                (byte)(voltage/AMPLIFIER_GAIN/5*NUMBER_OF_POTENTIOMETER_STEPS));
          return_code = RETURN_OK;
        }
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_SET_ACTUATION_FREQUENCY:
      if(payload_length()==sizeof(float)) {
        float freq = ReadFloat();
        // valid frequencies are 1kHz to 68MHz
        if(freq<1e3 || freq>68e6) {
          return_code = RETURN_GENERAL_ERROR;
        } else {
          uint8_t oct = 3.322*log(freq/1039)/log(10);
          uint16_t dac = round(2048-(2078*(float)(1<<(10+oct)))/freq);
          uint8_t cnf = 2; // CLK on, /CLK off
          // msb = OCT3 OCT2 OCT1 OCT0 DAC9 DAC8 DAC7 DAC6
          uint8_t msb = (oct << 4) | (dac >> 6);
          // lsb =  DAC5 DAC4 DAC3 DAC2 DAC1 DAC0 CNF1 CNF0
          uint8_t lsb = (dac << 2) | cnf;
          Wire.beginTransmission(LTC6904);
          Wire.write(msb);
          Wire.write(lsb);
          Wire.endTransmission();     // stop transmitting
          return_code = RETURN_OK;
        }
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_GET_STATE_OF_ELECTRODE:
      if(payload_length()==sizeof(uint8_t)+sizeof(uint16_t)) {
        uint16_t electrode = ReadUint16();
        if(electrode>=NUMBER_OF_ELECTRODES||electrode<0) {
          return_code = RETURN_BAD_INDEX;
        } else {
          Serialize(&electrode,sizeof(electrode));
          Serialize(&state_of_electrodes_[electrode],
                    sizeof(state_of_electrodes_[electrode]));
          return_code = RETURN_OK;
        }
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_GET_AVAILABLE_SAMPLING_RATES:
      if(payload_length()==0) {
        float sampling_rate;
        for(uint8_t i=7; i>0; i--) {
          sampling_rate = (float)16.0e6/(float)(1<<i)/(float)13;
          Serialize(&sampling_rate,sizeof(float));
        }
        return_code = RETURN_OK;
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;
    case CMD_SET_SAMPLING_RATE:
      if(payload_length()==sizeof(uint8_t)) {
        return_code = RETURN_OK;
        uint8_t adc_prescaler = 128>>ReadUint8();
        switch(adc_prescaler) {
          case 128:
            ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);
            break;
          case 64:
            ADCSRA |= _BV(ADPS2) | _BV(ADPS1);
            ADCSRA &= ~(_BV(ADPS0));
            break;
          case 32:
            ADCSRA |= _BV(ADPS2) | _BV(ADPS0);
            ADCSRA &= ~_BV(ADPS1);
            break;
          case 16:
            ADCSRA |= _BV(ADPS2);
            ADCSRA &= ~(_BV(ADPS1) | _BV(ADPS0));
            break;
          case 8:
            ADCSRA |= _BV(ADPS1) | _BV(ADPS0);
            ADCSRA &= ~_BV(ADPS2);
            break;
          case 4:
            ADCSRA |= _BV(ADPS1);
            ADCSRA &= ~(_BV(ADPS2) | _BV(ADPS0));
            break;
          case 2:
            ADCSRA |= _BV(ADPS0);
            ADCSRA &= ~(_BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0));
            break;
          default:
            return_code = RETURN_GENERAL_ERROR;
            break;
        }
      } else {
        return_code = RETURN_BAD_PACKET_SIZE;
      }
      break;

  case CMD_SAMPLE_PHASE:
       // sample the phase

            return_code = RETURN_OK;

   //         uint16_t delay_between_sets_ms = ReadUint16();
            for(uint8_t i=0; i<1; i++) {
              for(uint8_t j=0; j<1; j++) {
                for(uint8_t k=0; k<25; k++) {
                  phase_buffer_[i*1*1+j*1+k] = analogRead(3); //uint16_t phase_buffer_[2];
                }
              }
      /*        uint32_t t = millis();
              while(millis()-t<delay_between_sets_ms) {
              }*/
            }
/*
              for(uint8_t j=0; j<25; j++) {
                phase_buffer_[j] = analogRead(3); //which pin to read the phase
    //             uint32_t t = millis();
  //            while(millis()-t<delay_between_sets_ms) {
                   }
          //      }*/
              Serialize(phase_buffer_,25*sizeof(uint16_t));
            break;
            
    case CMD_SAMPLE_VOLTAGE:
      if(payload_length()<sizeof(uint8_t)*5+sizeof(uint16_t)) {
        return_code = RETURN_BAD_PACKET_SIZE;
      } else {
        uint8_t n_samples = ReadUint8();
        uint8_t n_sets = ReadUint8();
        uint16_t delay_between_sets_ms = ReadUint16();
        uint8_t n_channels = ReadUint8();
        if(n_samples*n_sets*n_channels>MAX_SAMPLES) {
          return_code = RETURN_GENERAL_ERROR;
        } else {
          if((payload_length()==3*sizeof(uint8_t)+sizeof(uint16_t)
             +n_channels*2*sizeof(uint8_t))
             || (payload_length()==3*sizeof(uint8_t)+sizeof(uint16_t)
             +n_channels*2*sizeof(uint8_t)
             +NUMBER_OF_ELECTRODES*sizeof(uint8_t))) {
            return_code = RETURN_OK;
            uint8_t channels[NUMBER_OF_AD_CHANNELS];
            uint8_t pot_steps[NUMBER_OF_AD_CHANNELS];
            for(uint8_t i=0; i<n_channels; i++) {
              channels[i] = ReadUint8();
              pot_steps[i] = ReadUint8();
              if(pot_steps[i]>NUMBER_OF_POTENTIOMETER_STEPS) {
                return_code = RETURN_GENERAL_ERROR;
                break;
              }
              // set the potentiometers
              switch(channels[i]) {
                case 0:
                  SendSPI(AD5206_SLAVE_SELECT_PIN,3,pot_steps[i]);
                  SendSPI(AD5206_SLAVE_SELECT_PIN,5,pot_steps[i]);
                  break;
                case 1:
                  SendSPI(AD5206_SLAVE_SELECT_PIN,0,pot_steps[i]);
                  SendSPI(AD5206_SLAVE_SELECT_PIN,1,pot_steps[i]);
                  break;
                default:
                  return_code = RETURN_GENERAL_ERROR;
                  break;
              }
            }
            // update the electrodes (if they were included in the packet)
            if(payload_length()==3*sizeof(uint8_t)+sizeof(uint16_t)
               +n_channels*2*sizeof(uint8_t)
               +NUMBER_OF_ELECTRODES*sizeof(uint8_t)){
              ReadArray(state_of_electrodes_,
                        NUMBER_OF_ELECTRODES*sizeof(uint8_t));
              UpdateAllElectrodes();
            }
            // sample the voltages
            for(uint8_t i=0; i<n_sets; i++) {
              for(uint8_t j=0; j<n_samples; j++) {
                for(uint8_t k=0; k<n_channels; k++) {
                  voltage_buffer_[i*n_samples*n_channels+j*n_channels+k] =
                    analogRead(channels[k]);
                }
              }
              uint32_t t = millis();
              while(millis()-t<delay_between_sets_ms) {
              }
            }
            Serialize(voltage_buffer_,
                      n_sets*n_samples*n_channels*sizeof(uint16_t));
          } else {
            return_code = RETURN_BAD_PACKET_SIZE;
          }
        }
      }
      break;
#endif
    default:
#ifndef AVR
      LogError("Unrecognized command", function_name);
#endif
      break;
  }
  SendReply(return_code);
}

void DmfController::ProcessReply(uint8_t cmd) {
  uint8_t reply_to = cmd^0x80;
#ifndef AVR
  const char* function_name = "ProcessReply()";
  sprintf(log_message_string_,
          "(0x%0X). This packet is a reply to command (%d)",
          reply_to,reply_to);
  LogMessage(log_message_string_, function_name);
  sprintf(log_message_string_,"Return code=%d",return_code());
  LogMessage(log_message_string_, function_name);
  sprintf(log_message_string_,"Payload length=%d",payload_length());
  LogMessage(log_message_string_, function_name);
#endif
  if(return_code()==RETURN_OK) {
#ifndef AVR // Replies that only the PC handles
    switch(reply_to) {
      case CMD_GET_DEVICE_NAME:
        LogMessage("CMD_GET_DEVICE_NAME", function_name);
        remote_name_ = ReadString();
        sprintf(log_message_string_,
                "remote_name_=%s",
                remote_name_.c_str());
        LogMessage(log_message_string_, function_name);
        break;
      case CMD_GET_DEVICE_VERSION:
        LogMessage("CMD_GET_DEVICE_VERSION", function_name);
        remote_version_ = ReadString();
        sprintf(log_message_string_,
                "remote_version_=%s",
                remote_version_.c_str());
        LogMessage(log_message_string_, function_name);
        break;
      case CMD_GET_NUMBER_OF_ELECTRODES:
        LogMessage("CMD_GET_DEVICE_VERSION", function_name);
        if(payload_length()==sizeof(uint16_t)) {
          state_of_electrodes_.resize(ReadUint16());
          sprintf(log_message_string_,
                  "state_of_electrodes_.size()=%d",
                  state_of_electrodes_.size());
          LogMessage(log_message_string_, function_name);
        } else {
          LogMessage("CMD_GET_NUMBER_OF_ELECTRODES, Bad packet size",
                     function_name);
        }
        break;
      case CMD_GET_STATE_OF_ALL_ELECTRODES:
        LogMessage("CMD_GET_STATE_OF_ALL_ELECTRODES", function_name);
        state_of_electrodes_.clear();
        for(int i=0; i<payload_length(); i++) {
          state_of_electrodes_.push_back(ReadUint8());
          sprintf(log_message_string_,
                  "state_of_electrodes_[%d]=%d",
                  i,state_of_electrodes_[i]);
          LogMessage(log_message_string_, function_name);
        }
        break;
      case CMD_SET_STATE_OF_ALL_ELECTRODES:
        LogMessage("CMD_SET_STATE_OF_ALL_ELECTRODES", function_name);
        LogMessage("all electrodes set successfully", function_name);
        break;
      case CMD_GET_STATE_OF_ELECTRODE:
        LogMessage("CMD_GET_STATE_OF_ELECTRODE", function_name);
        if(payload_length()==sizeof(uint16_t)+sizeof(uint8_t)) {
          uint16_t electrode = ReadUint16();
          if(electrode<state_of_electrodes_.size()) {
            state_of_electrodes_[electrode]=ReadUint8();
            sprintf(log_message_string_,
                  "electrode[%d]=%d",
                  electrode,state_of_electrodes_[electrode]);
            LogMessage(log_message_string_, function_name);
          } else {
            sprintf(log_message_string_,
                    "CMD_GET_STATE_OF_ELECTRODE, invalid elecrode (%d])",
                    electrode);
          }
          LogMessage(log_message_string_, function_name);
        } else {
            LogError("Bad packet size", function_name);
        }
        break;
      case CMD_SET_STATE_OF_ELECTRODE:
        LogMessage("CMD_SET_STATE_OF_ELECTRODE", function_name);
        LogMessage("electrode set successfully", function_name);
        break;
      case CMD_SET_ACTUATION_VOLTAGE:
      {
        LogMessage("CMD_SET_ACTUATION_VOLTAGE", function_name);
        LogMessage("volage set successfully", function_name);
        break;
      }
      case CMD_SET_ACTUATION_FREQUENCY:
      {
        LogMessage("CMD_SET_ACTUATION_FREQUENCY", function_name);
        LogMessage("frequency set successfully", function_name);
        break;
      }
      case CMD_SET_SAMPLING_RATE:
      {
        LogMessage("CMD_SET_SAMPLING_RATE", function_name);
        LogMessage("sampling rate set successfully", function_name);
        break;
      }
      case CMD_GET_AVAILABLE_SAMPLING_RATES:
      {
        LogMessage("CMD_GET_AVAILABLE_SAMPLING_RATES", function_name);
        available_sampling_rates_.clear();
        for(int i=0; i<payload_length()/sizeof(float); i++) {
          available_sampling_rates_.push_back(ReadFloat());
          sprintf(log_message_string_,"%.1f Hz",available_sampling_rates_[i]);
          LogMessage(log_message_string_,function_name);
        }
        break;
      }
      case CMD_SAMPLE_VOLTAGE:
      {
        LogMessage("CMD_SAMPLE_VOLTAGE", function_name);
        uint8_t n_samples = payload_length()/sizeof(uint16_t);
        sprintf(log_message_string_,"Read %d feedback samples",n_samples);
        LogMessage(log_message_string_,function_name);
        voltage_buffer_.resize(n_samples);
        for(int i=0; i<n_samples; i++) {
          voltage_buffer_[i] = ReadUint16();
        }
        break;
      }
      case CMD_SAMPLE_PHASE:
      {
        LogMessage("CMD_SAMPLE_PHASE", function_name);
        uint8_t n_samples_phase = payload_length()/sizeof(uint16_t);
        sprintf(log_message_string_,"Read %d phase samples",n_samples_phase);
        LogMessage(log_message_string_,function_name);
        phase_buffer_.resize(n_samples_phase);
        for(int i=0; i<n_samples_phase; i++) {
          phase_buffer_[i] = ReadUint16();
        }
        break;
      }
      default:
        LogError("Unrecognized command", function_name);
        break;
    }
#endif
  } else {
#ifndef AVR
    sprintf(log_message_string_,"return code=%d",
            return_code());
    LogError(log_message_string_, function_name);
#endif
  }
}

#ifdef AVR
////////////////////////////////////////////////////////////////////////////////
//
// These functions are only defined on the Arduino.
//
////////////////////////////////////////////////////////////////////////////////

void DmfController::begin() {
  pinMode(AD5206_SLAVE_SELECT_PIN, OUTPUT);
  Wire.begin();
  SPI.begin();

  // initialize the MAX7300
  // (code based on http://gdallaire.net/blog/?p=18)
  SendI2C(MAX7300_FIRST_ROW, 0x04, 0x01); // Shutdown disabled
  SendI2C(MAX7300_SECOND_ROW, 0x04, 0x01); // Shutdown disabled
  SendI2C(MAX7300_THIRD_ROW, 0x04, 0x01); // Shutdown disabled

  // set ports in output mode
  SendI2C(MAX7300_FIRST_ROW, 0x09, 0x55);
  SendI2C(MAX7300_FIRST_ROW, 0x0A, 0x55);
  SendI2C(MAX7300_FIRST_ROW, 0x0B, 0x55);
  SendI2C(MAX7300_FIRST_ROW, 0x0C, 0x55);
  SendI2C(MAX7300_FIRST_ROW, 0x0D, 0x55);
  SendI2C(MAX7300_FIRST_ROW, 0x0E, 0x55);
  SendI2C(MAX7300_FIRST_ROW, 0x0F, 0x55);

  SendI2C(MAX7300_SECOND_ROW, 0x09, 0x55);
  SendI2C(MAX7300_SECOND_ROW, 0x0A, 0x55);
  SendI2C(MAX7300_SECOND_ROW, 0x0B, 0x55); // P15, P14, P13, P12 ports in OUTPUT mode
  SendI2C(MAX7300_SECOND_ROW, 0x0C, 0x55);
  SendI2C(MAX7300_SECOND_ROW, 0x0D, 0x55);
  SendI2C(MAX7300_SECOND_ROW, 0x0E, 0x55);
  SendI2C(MAX7300_SECOND_ROW, 0x0F, 0x55);

  SendI2C(MAX7300_THIRD_ROW, 0x09, 0x55);
  SendI2C(MAX7300_THIRD_ROW, 0x0A, 0x55);
  SendI2C(MAX7300_THIRD_ROW, 0x0B, 0x55); // P15, P14, P13, P12 ports in OUTPUT mode
  SendI2C(MAX7300_THIRD_ROW, 0x0C, 0x55);
  SendI2C(MAX7300_THIRD_ROW, 0x0D, 0x55);
  SendI2C(MAX7300_THIRD_ROW, 0x0E, 0x55);
  SendI2C(MAX7300_THIRD_ROW, 0x0F, 0x55);

  // set all electrodes to ground
  UpdateAllElectrodes();

  // set all potentiometers to ground
  for(uint8_t i=1; i<7; i++) {
    SendSPI(AD5206_SLAVE_SELECT_PIN,i,0);
  }
  Serial.begin(DmfController::BAUD_RATE);
}

// send a command and some data to one of the I2C chips
// (code based on http://gdallaire.net/blog/?p=18)
void DmfController::SendI2C(uint8_t row, uint8_t cmd, uint8_t data) {
  Wire.beginTransmission(row);
  Wire.write(cmd);
  Wire.write(data);
  Wire.endTransmission();
}

void DmfController::SendSPI(uint8_t pin, uint8_t address, uint8_t data) {
  digitalWrite(pin,LOW);
  SPI.transfer(address);
  SPI.transfer(data);
  digitalWrite(pin,HIGH);
}

// update the state of all electrodes
void DmfController::UpdateAllElectrodes() {
  for(uint16_t i=0; i<NUMBER_OF_ELECTRODES; i++) {
    UpdateElectrode(i);
  }
}

// update the state of single electrode  //this must be fixed in the program*********
void DmfController::UpdateElectrode(const uint16_t electrode) {
  //set data byte
  uint8_t data = (state_of_electrodes_[electrode]>0); // right now, we only have 2 states (on/off)

  //set row address
  uint8_t row = MAX7300_FIRST_ROW;
  if(electrode>=24) row = MAX7300_SECOND_ROW;
  if(electrode>=52) row = MAX7300_THIRD_ROW;

  //set command byte
 // uint8_t cmd = MAX7300_PORT_ADDRESSES[electrode];
  uint8_t cmd;
  
  if (row == MAX7300_FIRST_ROW){
    cmd = MAX7300_PORT_ADDRESSES[electrode];}
  else if (row == MAX7300_SECOND_ROW && electrode < 48)
  { cmd = MAX7300_PORT_ADDRESSES[electrode%24];}
  else if (row == MAX7300_SECOND_ROW && electrode >= 48)
  { cmd = MAX7300_PORT_ADDRESSES[electrode%24+24];}
  else
  { cmd = MAX7300_PORT_ADDRESSES[electrode-52];}

  SendI2C(row, cmd, data);
}

#else
////////////////////////////////////////////////////////////////////////////////
//
// These functions are only defined on the PC.
//
////////////////////////////////////////////////////////////////////////////////

uint8_t DmfController::SendCommand(const uint8_t cmd) {
  const char* function_name = "SendCommand()";
  std::ostringstream msg;
  std::ostringstream vel;
  float ttime1 = MillisecondsSinceLast();
  msg << "time since last," << ttime1;
 // vel << "time since last," << ttime1;
  SerialProtocol::SendCommand(cmd);
  float ttime2 = MillisecondsSinceLast();
  msg << ",ms,returned in," << ttime2 << ",ms,command,"
      << (int)cmd << ",";
 // vel << ",ms,returned in," << ttime2 << ",ms,command,"
   //   << (int)cmd << ",";
  vel << "time since last," << ttime1 << ",ms,returned in," << ttime2 << ",ms,command,"
      << (int)cmd << "," << endl;
  if(return_code()!=RETURN_OK) {
    msg << "return code," << (int)return_code();
  }
  LogExperiment(msg.str().c_str());
  LogExperimentV(vel.str().c_str());
  msg.str("");
  vel.str("");
  msg << "returned " << (int)return_code();
  LogMessage(msg.str().c_str(), function_name);
  return return_code();
}

uint8_t DmfController::get_name(string& name) {
  const char* function_name = "get_name()";
  LogSeparator();
  LogMessage("send command", function_name);
  if(SendCommand(CMD_GET_DEVICE_NAME)==RETURN_OK) {
    name=remote_name_;
  }
  return return_code();
}

uint8_t DmfController::get_version(string& version) {
  const char* function_name = "get_version()";
  LogSeparator();
  LogMessage("send command", function_name);
  if(SendCommand(CMD_GET_DEVICE_VERSION)==RETURN_OK) {
    version=remote_version_;
  }
  return return_code();
}

uint8_t DmfController::get_number_of_electrodes(uint16_t& number_of_electrodes) {
  const char* function_name = "get_number_of_electrodes()";
  LogSeparator();
  LogMessage("send command", function_name);
  if(SendCommand(CMD_GET_NUMBER_OF_ELECTRODES)==RETURN_OK) {
    number_of_electrodes=state_of_electrodes_.size();
  }
  return return_code();
}

uint8_t DmfController::get_state_of_all_electrodes(vector<uint8_t>& state_of_electrodes) {
  const char* function_name = "get_state_of_all_electrodes()";
  LogSeparator();
  LogMessage("send command", function_name);
  if(SendCommand(CMD_GET_STATE_OF_ALL_ELECTRODES)==RETURN_OK) {
    state_of_electrodes=state_of_electrodes_;
  }
  return return_code();
};

uint8_t DmfController::get_state_of_electrode(const uint16_t electrode,uint8_t& state_of_electrode) {
  const char* function_name = "get_state_of_electrode()";
  LogSeparator();
  LogMessage("send command", function_name);
  Serialize(&electrode,sizeof(electrode));
  if(SendCommand(CMD_GET_STATE_OF_ELECTRODE)==RETURN_OK) {
    state_of_electrode=state_of_electrodes_[electrode];
  }
  return return_code();
};

uint8_t DmfController::get_available_sampling_rates(std::vector<float>& sampling_rates) {
  const char* function_name = "get_available_sampling_rates()";
  LogSeparator();
  LogMessage("send command", function_name);
  if(SendCommand(CMD_GET_AVAILABLE_SAMPLING_RATES)==RETURN_OK) {
    sampling_rates=available_sampling_rates_;
  }
  return return_code();
};

uint8_t DmfController::set_state_of_all_electrodes(const vector<uint8_t> state) {
  const char* function_name = "set_state_of_all_electrodes()";
  LogSeparator();
  LogMessage("send command", function_name);
  Serialize(&state[0],state.size()*sizeof(uint8_t));
  if(SendCommand(CMD_SET_STATE_OF_ALL_ELECTRODES)==RETURN_OK) {
    // check which electrodes have changed state
    vector<uint8_t> on,off;
    for(int i=0; i<state.size()&&i<state_of_electrodes_.size(); i++) {
      if(state[i]!=state_of_electrodes_[i]) {
        if(state[i]==0) {
          off.push_back(i);
        } else {
          on.push_back(i);
        }
      }
    }
    // update state
    state_of_electrodes_=state;

    // log experiment
    std::ostringstream msg;
    msg << "set_state_of_all_electrodes,";
    if(on.size()||off.size()) {
      if(on.size()>0) {
        msg << "Turn on:,";
        for(int i=0; i<on.size(); i++) {
          msg << (int)on[i] << ",";
        }
      }
      if(off.size()>0) {
        msg << "Turn off:,";
        for(int i=0; i<off.size(); i++) {
          msg << (int)off[i] << ",";
        }
      }
    }
    msg << endl;
    LogExperiment(msg.str().c_str());
  }
  return return_code();
}

uint8_t DmfController::set_state_of_electrode(const uint16_t electrode, const uint8_t state) {
  const char* function_name = "set_state_of_electrode()";
  LogSeparator();
  LogMessage("send command", function_name);
  Serialize(&electrode,sizeof(electrode));
  Serialize(&state,sizeof(state));
  if(SendCommand(CMD_SET_STATE_OF_ELECTRODE)==RETURN_OK) {
    // keep track of the electrode state
    if(electrode>state_of_electrodes_.size()-1) {
      state_of_electrodes_.resize(electrode);
    }
    state_of_electrodes_[electrode]=state;

    // log experiment
    std::ostringstream msg;
    std::ostringstream vel;
    msg << "set_state_of_electrode,";
    if(state==0) {
      msg << "Turn off:," << electrode << endl;
    } else {
      msg << "Turn on:," << electrode << endl;
      vel << "Turn on:," << electrode << endl;
    }
    LogExperiment(msg.str().c_str());
    LogExperimentV(vel.str().c_str());
  }
  return return_code();
}

uint8_t DmfController::set_actuation_voltage(const float v_rms){
  const char* function_name = "set_actuation_voltage()";
  LogSeparator();
  LogMessage("send command", function_name);
  Serialize(&v_rms,sizeof(v_rms));
  if(SendCommand(CMD_SET_ACTUATION_VOLTAGE)==RETURN_OK) {
    std::ostringstream msg;
    msg << "set_actuation_voltage," << v_rms << ",Vrms" << endl;
    LogExperiment(msg.str().c_str());
  }
  return return_code();
}

uint8_t DmfController::set_actuation_frequency(const float freq_hz) {
  const char* function_name = "set_actuation_frequency()";
  LogSeparator();
  sprintf(log_message_string_,"freq_hz=%.1f",freq_hz);
  LogMessage(log_message_string_, function_name);
  LogMessage("send command", function_name);
  Serialize(&freq_hz,sizeof(freq_hz));
  if(SendCommand(CMD_SET_ACTUATION_FREQUENCY)==RETURN_OK) {
    std::ostringstream msg;
    msg << "set_actuation_frequency," << freq_hz/1000 << ",kHz" << endl;
    LogExperiment(msg.str().c_str());
  }
  return return_code();
}

uint8_t DmfController::set_sampling_rate(const uint8_t sampling_rate) {
  const char* function_name = "set_sampling_rate()";
  LogSeparator();
  LogMessage("send command", function_name);
  Serialize(&sampling_rate,sizeof(sampling_rate));
  if(SendCommand(CMD_SET_SAMPLING_RATE)==RETURN_OK) {
    std::ostringstream msg;
    msg << "set sampling rate," << (int)sampling_rate << endl;
    LogExperiment(msg.str().c_str());
  }
  return return_code();
}

uint8_t DmfController::SampleVoltage(std::vector<uint16_t>& voltage_buffer,
                                     std::vector<uint8_t> ad_channels,
                                     std::vector<uint8_t> pot_steps,
                                     uint8_t n_samples,
                                     uint8_t n_sets,
                                     uint16_t delay_between_sets_ms,
                                     const std::vector<uint8_t> state) {
  const char* function_name = "SampleVoltage()";
  LogSeparator();
  LogMessage("send command", function_name);

  if(n_samples>MAX_SAMPLES*ad_channels.size()*n_sets
     || ad_channels.size()>NUMBER_OF_AD_CHANNELS) {
    return RETURN_GENERAL_ERROR;
  }
  for(uint8_t i=0; i<pot_steps.size(); i++) {
    if(pot_steps[i]>NUMBER_OF_POTENTIOMETER_STEPS) {
      return RETURN_GENERAL_ERROR;
    }
  }
  // if we get this far, everything is ok
  Serialize(&n_samples,sizeof(n_samples));
  Serialize(&n_sets,sizeof(n_sets));
  Serialize(&delay_between_sets_ms,sizeof(delay_between_sets_ms));
  uint8_t n_channels = ad_channels.size();
  Serialize(&n_channels,sizeof(n_channels));
  for(uint8_t i=0; i<ad_channels.size(); i++) {
    Serialize(&ad_channels[i],sizeof(ad_channels[i]));
    Serialize(&pot_steps[i],sizeof(pot_steps[i]));
  }
  std::ostringstream msg;
  std::ostringstream vel;
  msg << "SampleVoltage,";

  // if we're updating the electrode state
  if(state.size()) {
    Serialize(&state[0],state.size()*sizeof(uint8_t));

    // check which electrodes have changed state
    vector<uint8_t> on,off;
    for(int i=0; i<state.size()&&i<state_of_electrodes_.size(); i++) {
      if(state[i]!=state_of_electrodes_[i]) {
        if(state[i]==0) {
          off.push_back(i);
        } else {
          on.push_back(i);
        }
      }
    }
    // update state
    state_of_electrodes_=state;

    if(on.size()||off.size()) {
      if(on.size()>0) {
        msg << "Turn on:,";
        vel << "Turn on:,";
        for(int i=0; i<on.size(); i++) {
          msg << (int)on[i] << ",";
          vel << (int)on[i] << "," << endl;
        }
      }
      if(off.size()>0) {
        msg << "Turn off:,";
        for(int i=0; i<off.size(); i++) {
          msg << (int)off[i] << ",";
        }
      }
    }
  }
  msg << endl << CSV_INDENT_ << "n_samples,"
      << (int)n_samples << endl << CSV_INDENT_ << "n_SendSPIsets," << (int)n_sets
      << endl << CSV_INDENT_ << "delay_between_sets_ms,"
      << (int)delay_between_sets_ms << endl;
  if(SendCommand(CMD_SAMPLE_VOLTAGE)==RETURN_OK) {
    voltage_buffer = voltage_buffer_;
    for(int i=0; i<ad_channels.size(); i++) {
      msg << CSV_INDENT_ << "pot," << (float)pot_steps[0]/(float)NUMBER_OF_POTENTIOMETER_STEPS
                         *(float)POTENTIOMETER_MAX << endl;
      for(int j=0; j<n_sets; j++) {
        msg << CSV_INDENT_ << "voltage_buffer[" << i << "][" << j << "],";
        // calculate the DC bias
        double dc_bias = 0;
        for(int k=0; k<n_samples; k++) {
          dc_bias += (float)voltage_buffer[j*ad_channels.size()*n_samples+
                                           k*ad_channels.size()+i]
                                           /1024*5/n_samples;
          msg << (float)voltage_buffer[j*ad_channels.size()*n_samples+
                                       k*ad_channels.size()+i]/1024*5 << ",";
        }
        double v_rms = 0;
        for(int k=0; k<n_samples; k++) {
          v_rms += pow((float)voltage_buffer[j*ad_channels.size()*n_samples+
                                             k*ad_channels.size()+i]
                                             /1024*5-dc_bias,2)/n_samples;
        }
        v_rms = sqrt(v_rms);
        msg << endl << CSV_INDENT_ << "dc_bias," << dc_bias << endl
            << CSV_INDENT_ << "v_rms," << v_rms << endl;
        vel << ",,,,,,,v_rms," << v_rms << endl;
      }
    }
    LogExperiment(msg.str().c_str());
    LogExperimentV(vel.str().c_str());
  }
  return return_code();
}


uint8_t DmfController::SamplePhase(std::vector<uint16_t>& phase_buffer) {
  const char* function_name = "SamplePhase()";
  LogSeparator();
  LogMessage("send command", function_name);
  //Serialize(2,sizeof(2));
  if(SendCommand(CMD_SAMPLE_PHASE)==RETURN_OK) {
    phase_buffer = phase_buffer_;
  }
  return return_code();
}

uint8_t DmfController::SetExperimentLogFile(const char* file_name) {
    char* function_name = "SetExperimentLogFile()";
  if(experiment_log_file_.is_open()) {
    vel_file.close();
    experiment_log_file_.close();
  }
  std::ostringstream msg;
  msg << "file_name=" << file_name;
  LogMessage(msg.str().c_str(), function_name);
  experiment_log_file_name_ = file_name;
  vel_file_name_ = "velocity.csv";
  t_last_ = boost::posix_time::microsec_clock::universal_time();
  vel_file.open(vel_file_name_.c_str(), ios::app);
  experiment_log_file_.open(experiment_log_file_name_.c_str(), ios::app);

  if(experiment_log_file_.fail()==false) {
    return RETURN_OK;
  } else {
    return RETURN_GENERAL_ERROR;
  }
}






void DmfController::LogExperiment(const char* msg) {
if(experiment_log_file_.is_open()) {
    experiment_log_file_ << msg;
    experiment_log_file_.flush();
  }
}

void DmfController::LogExperimentV(const char* vel) {

  if(vel_file.is_open()) {
    vel_file << vel;
    vel_file.flush();
  }
}

float DmfController::MillisecondsSinceLast() {
  boost::posix_time::ptime t = t_last_;
  t_last_ = boost::posix_time::microsec_clock::universal_time();
  return round((float)(boost::posix_time::microsec_clock::universal_time()-t)
         .total_microseconds()/1000);
}

#endif // AVR
