////////////////////////////////////////////////////////////////////////////////
//
// SerialProtocol
//
// This class implements a simple communication protocol between a PC and
// Arduino over an RS-232 link using HDLC-like framing.  We use a wrapper
// for the boost asio library (SimpleSerial) that provides an interface that
// closely matches the Serial library on the Arduino.  This allows us to share
// the bulk of the code between the PC (Windows, Linux or Mac) and the Arduino.
//
// Each packet has the following structure:
//
//  +------------+----------+----------+-------------+------------+----------+
//  | Start Flag | Command  |  Length  |   Payload   |  Checksum  | End Flag |
//  |   1 byte   |  1 byte  |  1 byte  | 0 to Length |  2 bytes   |  1 byte  |
//  |    0x7E    |          |          |             | (optional) |   0x7E   |
//  +------------+----------+----------+-------------+------------+----------+
//
// Total packet length (not including flags) = Length + 4 (with checksum)
//                                             Length + 2 (without checksum)
//
// To use this class, you must derive a class based on it and reimplement the
// virtual member function "ProcessPacket(...)".
//
////////////////////////////////////////////////////////////////////////////////

#ifndef _SERIAL_PROTOCOL_H
#define	_SERIAL_PROTOCOL_H

#include <stdint.h>

#ifndef AVR
  #include "logging.h"
  #include "SimpleSerial.h"
  #include <string>
#endif

class SerialProtocol {
public:
#ifndef AVR
  static const uint32_t TIMEOUT_MICROSECONDS = 3000000; // TODO: this should be configurable
#endif

  // protocol constants
  static const uint8_t MAX_PAYLOAD_LENGTH =       255;
  static const uint8_t MAX_STRING_SIZE =          160;

  // reserved commands
  static const uint8_t CMD_GET_PROTOCOL_NAME =    0x80;
  static const uint8_t CMD_GET_PROTOCOL_VERSION = 0x81;

  // reserved return codes
  static const uint8_t RETURN_OK =                0x00;
  static const uint8_t RETURN_GENERAL_ERROR =     0x01;
  static const uint8_t RETURN_UNKNOWN_COMMAND =   0x02;
  static const uint8_t RETURN_TIMEOUT =           0x03;
  static const uint8_t RETURN_NOT_CONNECTED =     0x04;
  static const uint8_t RETURN_BAD_INDEX =         0x05;
  static const uint8_t RETURN_BAD_PACKET_SIZE =   0x06;

  SerialProtocol(uint32_t baud_rate,
                 bool checksum_enabled_
#ifndef AVR
                 ,const char* class_name
#endif
                 );
  ~SerialProtocol();

  void set_debug(const bool debug);

#ifdef AVR
  void Listen();
#else
  // these methods query the remote device for protocol name/version
  uint8_t get_protocol_name(std::string& name);
  uint8_t get_protocol_version(std::string& version);
  uint8_t Connect(const char* port);
#endif

  // these methods force the derived class to define a protocol name and version
  virtual const char* protocol_name() = 0;
  virtual const char* protocol_version() = 0;

protected:
  // these virtual methods must be overriden in the derived class
  virtual void ProcessCommand(uint8_t cmd) = 0;
  virtual void ProcessReply(uint8_t cmd) = 0;
  uint8_t return_code() {return return_code_; }
  uint8_t payload_length() { return payload_length_; }

  template<typename T>
    void Serialize(T data,uint8_t size) {
      Serialize((const uint8_t*)data,size); }
  void Serialize(const uint8_t* u, const uint8_t size);
  void SendReply(const uint8_t return_code);

  template<typename T> void ReadArray(T* array, const uint8_t size) {
#ifndef AVR
    LogMessage("","ReadArray()");
#endif
    bytes_read_ += size;
    memcpy(array,payload_+bytes_read_-size,size);
  }
  const char* ReadString();
  uint16_t ReadUint16();
  uint8_t ReadUint8();
  float ReadFloat();
  uint8_t WaitForReply();
  uint8_t SendCommand(const uint8_t cmd);

#ifndef AVR
  inline void LogMessage(const char* msg,
                         const char* function_name,
                         uint8_t level=5) {
          if(debug_) {
            Logging::LogMessage(level,msg,class_name_.c_str(),function_name); }}
  inline void LogError(const char* msg, const char* function_name) {
          if(debug_) {
          Logging::LogError(msg,class_name_.c_str(),function_name); }}
  inline void LogSeparator() { if(debug_) { Logging::LogSeparator(); }}
  static char log_message_string_[];
#endif

private:
  uint8_t packet_cmd_; // command
  uint8_t payload_[MAX_PAYLOAD_LENGTH]; // payload
  uint8_t payload_length_; // length of the packet we are reading
  uint8_t return_code_; // return code
  void SendPreamble();
  void SendPayload();
  void SendByte(uint8_t b);
  static const uint8_t FRAME_BOUNDARY =           0x7E;
  static const uint8_t CONTROL_ESCAPE =           0x7D;
  static const uint8_t ESCAPE_XOR =               0x20;

  void ProcessPacket();
  void ProcessSerialInput(const uint8_t byte);
  uint32_t baud_rate_;
  uint8_t bytes_received_; // bytes received so far in packet
  uint8_t bytes_read_; // bytes that have been read (by Read methods)
  uint8_t bytes_written_; // bytes that have been written (by Serialize method)
  bool un_escaping_; // flag that the last byte was an escape
  bool waiting_for_reply_; // flag that we are waiting for a response
  bool checksum_enabled_;
  bool debug_;
#ifndef AVR
  SimpleSerial Serial;
  std::string class_name_;
  boost::posix_time::ptime time_cmd_sent_;
#endif
};

#endif	// _SERIAL_PROTOCOL_H
