/*****************
 * "electrode.hpp"
 *
 *  Copyright 2014
 *
 *  This file is part of Actuation.
 *
 *  Actuation is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Actuation is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Actuation.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Contains include statements, the definitions of the point struct and electrode class,
 *  and function definitions for "electrode.cpp"
 *
 *****************/


#ifndef _ELECTRODE_H_
#define _ELECTRODE_H_

#include "include.h"
#include "path.h"

//TODO: update code to use these defines
#define CONNECTOR 2
#define ELECTRODE 1
#define WIRE 0

//Electrode States
#define HIGH 1
#define GROUND 0

//Row
#define BOTTOM 0
#define TOP 1

extern Config configuration;

//check whether two line segments intersect
int line_intersection(point p1, point q1, point p2, point q2);
int prec_line_intersection(point p1, point q1, point p2, point q2);//extends the endpoints slightly
int equal(double a, double b);

class Electrode {
private:
  std::vector<Path*> shapes;
  int type;                 //CONNECTOR, WIRE, ELECTRODE
  int position;             //1 for top, 0 for bottom

  //electrodes may be connected to several connectors
  std::set<int> numberSet;     //1-80
  std::set<int> indexSet;

  //GROUND - 0, HIGH VOLTAGE - 1
  int* state;

public:
  //used for sorting electrodes based on position
  Electrode* prev;
  Electrode* next;

  int get_type(){return type;};
  int get_state(int step){return state[step];};
  int getFirstNumber(){return *(numberSet.begin());};
  int getNumber(int nIndex) {
    set<int>::iterator it = numberSet.begin();
    int sIndex = 0;
    for(;it!=numberSet.end();it++, sIndex++) {
      if(sIndex == nIndex) { return (*it); }
    }
    return -1;
  };
  int getIndex(int i){
    std::set<int>::iterator it = indexSet.begin();
    int sIndex = 0;
    for(; it != indexSet.end(); it++, sIndex++) {
      if(sIndex == i) {return (*it); }
    }
    return -1;
  };
  void purge() {
    set<int>::iterator it = numberSet.begin();
    for(;it!=numberSet.end();it++) {
      //if((*it) > 40) { numberSet.erase(it); }
      if((*it) > 80) { numberSet.erase(it); }
    }
  };

  void setNumberString(std::string nString) {
    numberSet.clear();
    if(nString.length() == 0) {
      return;
    }
    size_t firstComma = nString.find(',', 0);
    size_t secondComma = nString.find(',',firstComma+1);

    int currentNumber = atoi(nString.substr(0, firstComma).data());
    addNumber(currentNumber);

    while (firstComma != nString.rfind(",")) {
      secondComma = nString.find(',', firstComma+1);
      currentNumber = atoi(nString.substr(firstComma+1,
                                          secondComma-firstComma-1).data());
      addNumber(currentNumber);
      firstComma = secondComma;
    }
    currentNumber = atoi(nString.substr(firstComma + 1, nString.length()).data());
    addNumber(currentNumber);
  };

  void incrementRow(int row, int increment) {
    set<int>::iterator it = numberSet.begin();
    int cNumber;
    for(;it!=numberSet.end();it++) {
      switch(row)
        case TOP: {
       //   if((*it)>20) {
            if((*it)>40) {
            cNumber = (*it) + increment;
            numberSet.erase(it);
            addNumber(cNumber);
          }
          break;
        case BOTTOM:
          //if((*it)<21) {
            if((*it)<41) {
            cNumber = (*it) + increment;
            numberSet.erase(it);
            addNumber(cNumber);
          }
          break;
        default:
          break;
      }
    }
  };

  int size(){return shapes.size();};
  int numberSetSize(){return numberSet.size();};
  int indexSetSize(){return indexSet.size();};

  bool isEmpty(){return numberSetSize()==0;};
  bool isInSet(int num){return numberSet.find(num) != numberSet.end();};
  int get_position(){return position;};
  void set_type(int _type){type = _type;};
  void set_position(int _position){position = _position;};
  void addNumber(int number){numberSet.insert(number);};
  void addIndex(int index){indexSet.insert(index);};
  bool addSet(std::set<int>::iterator begin, std::set<int>::iterator end){
    unsigned int oldSize = indexSet.size();
    indexSet.insert(begin, end);
    return !(oldSize==indexSet.size());
  };
  std::set<int>::iterator begin(){return indexSet.begin();};
  std::set<int>::iterator end(){return indexSet.end();};

  void setState(int step, int _state){state[step] = _state;};

  point getPoint(int pi){return shapes[pi]->getPoint(); };
  point frontPoint(){return shapes.front()->getPoint();};
  point backPoint(){return shapes.back()->getPoint();};

  //void increment(int inc){number+=inc;};
  void clear_step(int step){state[step]=GROUND;};
  void clear() {
    for(int i = 0; i < configuration.max_steps(); i++) {
      state[i] = GROUND;
    }
  };
  void copy_step(int source, int dest) {
    state[dest] = state[source];
  };

  //create an electrode given an input string
  Electrode(char* p);
  Electrode();
  ~Electrode();

  //add a new electrode to a sorted linked list of electrodes
  void addElectrode(Electrode* n);

  //check whether a point is within a electrode
  int check_connection(point p);

  //check whether two wires intersect at any of their line segments
  int check_wire_intersection(Electrode* line2);

  //check whether a line and a polygon intersect
  int check_intersection(point p, point q);

  //change the state of this electrode at the current step
  void changeState(int current_step);

  std::string getNumberString();
  std::string getCharString();

  void draw(cairo_t* cr, double scale, point top_left, int cState);
  void draw_selected(cairo_t* cr, double scale, point top_left);
  void print();

  point find_top_left();
  point find_bottom_right();
  point find_bottom_left();
  point find_center();
  bool endpoints_match();

  void multiply(int multiple) {
    for(int i = 0; i < size(); i++) {
      shapes[i]->multiply(multiple);
    }
  };
};

#endif
