/*****************
 * "shape.hpp"
 *
 *  Copyright 2010 Andrei Vovk
 *
 *  This file is part of Actuation.
 *
 *  Actuation is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Actuation is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Actuation.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  Description:
 *
 *  Code for the Shape class and its children: Path, Curve, and Arc
 *
 *
 *
 *
 *****************/

#ifndef _PATH_H_
#define _PATH_H_

#include "include.h"

extern Config configuration;

using namespace std;

struct point{
  double x;
  double y;
};

double cross_mult(point u, point v);

//defined by an endpoint
class Path{
public:
  double getX(){return p.x;};
  double getY(){return p.y;};
  point getPoint(){return p;};
  char getType(){return type;};
  virtual void draw(cairo_t * cr, double scale, point top_left)
  {
    cairo_line_to(cr, scale * (p.x - top_left.x), scale * (p.y - top_left.y));
  };
  virtual void multiply(int multiple)
  {
    p.x *= multiple;
    p.y *= multiple;
  };
protected:
  point p;
  char type;

};

/* An elliptical Arc
 *
 * draw() does not work correctly
 *
 * Cairo defines an arc using the center, radius, and starting and ending angles
 * SVG defines an arc using the starting and ending points, radii, x-axis rotation,
 * and the large arc and sweep flags
 */
class Arc : public Path
{
  point radius;
  point prev;
  point center;
  double start_angle;
  double end_angle;
  double rotation;
  int large_arc;
  int sweep;
public:
  //lowercase
  Arc(double radx, double rady, int flag1, int flag2, int flag3, double endx, double endy, Path * previous){
    type = 'a';
    radius.x = radx;
    radius.y = rady;
    large_arc=flag2;
    sweep=flag3;
    rotation = flag1;
    p.x = endx + previous->getX();
    p.y = endy + previous->getY();
    prev.x = previous->getX();
    prev.y = previous->getY();
    //compute();
    simple();
  };
  //uppercase
  Arc(double radx, double rady, int flag1, int flag2, int flag3, double endx, double endy){
    type = 'a';
    radius.x = radx;
    radius.y = rady;
    large_arc=flag2;
    sweep=flag3;
    rotation=flag1;
    p.x = endx;
    p.y = endy;
    //compute();
    simple();
  };
  void simple()
  {
    center.x = prev.x + p.x;
    center.x /= 2;
    center.y = prev.y + p.y;
    center.y /=2;
    start_angle = 0;
    end_angle = 0;


  }

  /*convert from the arc description given by the svg to those required by cairo
   * SVG gives: prev(start point), radius, angle of rotation, large arc flag, sweep flag, and p (end point)
   * Cairo requires: center, radius, start angle, end angle
   * Code based on: http://www.w3.org/TR/SVG/implnote.html#ArcConversionEndpointToCenter
   *
   *
   */
  void compute()
  {
    point dprev;
    point dcenter;
    if(radius.x == 0 || radius.y ==0){
      radius.x = 0;
      simple();
      return;
    }
    dprev.x = cos(rotation) * (prev.x - p.x)/2 + sin(rotation) * (prev.y - p.y)/2;
    dprev.y = -sin(rotation) * (prev.x - p.x)/2 + cos(rotation) * (prev.y - p.y)/2;
    double coef;
    coef = ((radius.x * radius.x)*(radius.y * radius.y)
	    -(radius.x * radius.x)*(dprev.y * dprev.y)
	    -(radius.y * radius.y)*(dprev.x * dprev.x))/
      ((radius.x * radius.x)*(dprev.y * dprev.y)
       +(radius.y * radius.y)*(dprev.x * dprev.x));
    if(coef < 0)
      coef *= -1;
    coef = sqrt(coef);
    if(large_arc == sweep)
      coef *= -1;
    dcenter.x = coef * radius.x * dprev.y / radius.y;
    dcenter.y = coef * (-radius.y) * dprev.x / radius.x;

    center.x = cos(rotation) * dcenter.x - sin(rotation) * dcenter.y
      + (prev.x + p.x)/2;
    center.y = sin(rotation) * dcenter.x + cos(rotation) * dcenter.y
      + (prev.y + p.y)/2;

    point u, v;
    u.x = 1;
    u.y = 0;
    v.x = (dprev.x - dcenter.x)/radius.x;
    v.y = (dprev.y - dcenter.y)/radius.y;

    start_angle = acos(cross_mult(u, v));
    if(u.x * v.y - u.y * v.x < 0)
      start_angle *= -1;
    while(start_angle > 360)
      start_angle -= 360;
    while(start_angle < -360)
      start_angle += 360;
    u.x = (dprev.x - dcenter.x)/radius.x;
    u.y = (dprev.y - dcenter.y)/radius.y;
    v.x = (-dprev.x - dcenter.x)/radius.x;
    v.y = (-dprev.y - dcenter.y)/radius.y;
    end_angle = ((int)acos(cross_mult(u, v)))%360;
    if(u.x * v.y - u.y * v.x < 0)
      end_angle *= -1;
    if(sweep == 0 && end_angle > 0)
      end_angle -= 360;
    else if (sweep == 1 && end_angle < 0)
      end_angle += 360;


  };

  //draw a circular arc
  //by approximating as a bezier curve
  //Note: in SVG, an elliptical arc is described, but the radii have been the same in all tested files
  //so radius.x is used
  //there is a way to draw an elliptical arc in cairo, an example is here:
  // http://cairographics.org/manual/cairo-paths.html#cairo-arc
  /* void draw(cairo_t * cr, double scale, point top_left)
     {
     if(!sweep)
     cairo_arc(cr, scale * (center.x - top_left.x), scale * (center.y - top_left.y), radius.x, start_angle, end_angle);
     else
     cairo_arc_negative(cr, scale * (center.x - top_left.x), scale * (center.y - top_left.y), radius.x, start_angle, end_angle);


     //approximate using a cubic bezier curve

     double theta = atan(((double)(p.y-prev.y) )/((double)(p.x-prev.x) ));
     double dy = radius.x * sin(theta);
     double dx = radius.x * cos(theta);
     //theta is undefined (x values are equal)
     if(theta!=theta){
     dy = 0;
     dx = radius.x;
     }

     point bcontrol, econtrol;
     bcontrol.x = prev.x + dx;
     bcontrol.y = prev.y + dy;
     econtrol.x = p.x + dx;
     econtrol.y = p.y + dy;
     cout<<"curve: "<<bcontrol.x<<"y"<<bcontrol.y<<" "<<econtrol.x<<" "<<econtrol.y<<endl;
     cairo_curve_to(cr,scale*(bcontrol.x - top_left.x) ,scale*(bcontrol.y-top_left.y) ,scale*(econtrol.x-top_left.x), scale*(econtrol.y-top_left.y), scale*(p.x - top_left.x), scale*(p.y- top_left.y));


     }*/
};

class Curve : public Path
{
private:
  point bcontrol;
  point econtrol;
public:
  //lowercase
  Curve(double bcontrolx, double bcontroly, double econtrolx, double econtroly, double endx, double endy, Path * previous)
  {

    type = 'c';
    bcontrol.x = bcontrolx + previous->getX();
    bcontrol.y = bcontroly + previous->getY();
    econtrol.x = econtrolx + previous->getX();
    econtrol.y = econtroly + previous->getY();;
    p.x = endx + previous->getX();
    p.y = endy + previous->getY();
  }
  point getB(){return bcontrol;};
  point getE(){return econtrol;};
  void draw(cairo_t * cr, double scale, point top_left)
  {
    cairo_curve_to(cr,scale*(bcontrol.x - top_left.x) ,scale*(bcontrol.y-top_left.y) ,scale*(econtrol.x-top_left.x), scale*(econtrol.y-top_left.y), scale*(p.x - top_left.x), scale*(p.y- top_left.y));
  };
  void multiply(int multiple)
  {
    p.x *= multiple;
    p.y *= multiple;
    bcontrol.x *= multiple;
    bcontrol.y *= multiple;
    econtrol.x *= multiple;
    econtrol.y *= multiple;
  }
};

//A Line is the simplest of Paths, defined only by a point.
//It is equivalent to a straight line drawn from the endpoint
//of the previous Shape, to the current.
class Line : public Path
{
public:
  //uppercase or first
  Line(double x, double y){ p.x = x; p.y = y;type = 'p';};
  //lowercase (relative coordinates)
  Line(double x, double y, Path * previous){ p.x = previous->getX() + x; p.y = previous->getY() + y; type='p';};
  //draw a straight line
  void draw(cairo_t * cr, double scale, point top_left)
  {
    cairo_line_to(cr, scale * (p.x - top_left.x), scale * (p.y - top_left.y));
  };
};

#endif
